"""Common definitions for all test sequencies"""
import os

# This DUT definition is used only if SN input comes from UI.
# Otherwise DUT definition must come from prepare_test()
DUTS = ["left", "right", "middle"]

# Wait DUT serial number from UI
SN_FROM_UI = True

# List of test sequences' directory names
TEST_SEQUENCES = os.listdir('test_definitions')
if 'common' in TEST_SEQUENCES:
    TEST_SEQUENCES.remove('common')



#! /usr/bin/python
"""
Test sequencer for any production testing
"""

import os
import argparse
import sys
import threading
from queue import Queue
from distutils.dir_util import copy_tree
from test_runner import runner
import listener.listener as listener
import logging
import tornado
import json
import pathlib
import yaml
import coloredlogs
import logging.config

PORT = 4321

PARSER = argparse.ArgumentParser(description="Iris test sequencer.")
PARSER.add_argument("--single_run", "-s", help="Run only once", action="store_true")
PARSER.add_argument(
    "--create",
    "-c",
    metavar='path',
    type=str,
    nargs=1,
    help="Creates empty/example test COMMON_definitions",
)
PARSER.add_argument("--report_off", "-r", help="Don't create test report", action="store_true")
PARSER.add_argument(
    "--listener",
    "-l",
    help="Creates HTTP listener. Testing is then started through REST API.",
    action="store_true",
)


PARSER.add_argument(
    "-v",
    "--verbose",
    help="Increase output verbosity. Sets colored console logging to debug level.",
    action="store_true",
)

PARSER.add_argument(
    '--no_colored_logs', '-n', help='Disables colored logs on console.', action="store_true"
)

PARSER.add_argument(
    "--list-applications",
    "-a",
    help="Lists available application on the connected Gaia tester (G5 or other). Gaia instrument must be defined and available.",
    action="store_true",
)

PARSER.add_argument('-p', '--port', help="Set port to listen", type=int)

ARGS = PARSER.parse_args()


LOG_SETTINGS_FILE = pathlib.Path('test_definitions/common/logging.yaml')


if LOG_SETTINGS_FILE.is_file():
    with LOG_SETTINGS_FILE.open() as _f:
        LOG_CONF = yaml.safe_load(_f.read())
    pathlib.Path(LOG_CONF['log_file_path']).mkdir(parents=True, exist_ok=True)
    logging.config.dictConfig(LOG_CONF)
    logging.info('Logging with configuration from %s', LOG_SETTINGS_FILE)
else:
    logging.basicConfig(level=logging.INFO)
    logging.warning('Cannot find logging settings. Logging with basicConfig.')

if not ARGS.no_colored_logs:
    HANDLERS = logging.getLogger().handlers
    CONSOLE_LOG_LEVEL = list(
        filter(lambda x: x and x.name and x.name.lower() == 'console', HANDLERS)
    )

    if CONSOLE_LOG_LEVEL:
        CONSOLE_LOG_LEVEL = CONSOLE_LOG_LEVEL[0].level
        if ARGS.verbose:
            CONSOLE_LOG_LEVEL = logging.DEBUG

        coloredlogs.install(level=CONSOLE_LOG_LEVEL, milliseconds=True)

LOGGER = logging.getLogger(__name__)

LOGGER.debug("Logging initialized")

if ARGS.create:

    from test_definition_template import example_sequence

    TEST_DEF_PATH = os.path.join("./test_definitions", ARGS.create[0])

    if os.path.isdir(TEST_DEF_PATH):
        print("test_definitions folder already exists")
        sys.exit(-1)

    copy_tree(example_sequence.__path__[0], TEST_DEF_PATH)

    if not os.path.isdir('./test_definitions/common'):
        from test_definition_template import common

        copy_tree(common.__path__[0], './test_definitions/common')

        import additional_dist_files
        from shutil import copyfile
        copyfile(additional_dist_files.__path__[0] + '/Dockerfile', './Dockerfile')
    else:
        print('./test_definitions/common already exists. Not copying it.')

    sys.exit(0)

CONTROL = runner.get_test_control()
COMMON_DEFINITIONS = runner.get_common_definitions()

if ARGS.list_applications:
    # Print available applications and actions
    class GaiaJsonEncoder(json.JSONEncoder):
        '''Encode json properly'''

        def default(self, obj):
            if callable(obj):
                return obj.__name__
            # Let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, obj)

    COMMON_DEFINITIONS.instrument_initialization()

    client = COMMON_DEFINITIONS.INSTRUMENTS['gaia']

    print(json.dumps(client.applications, indent=4, sort_keys=True, cls=GaiaJsonEncoder))
    print(json.dumps(client.state_triggers, indent=4, sort_keys=True, cls=GaiaJsonEncoder))
    sys.exit()

CONTROL['run'].set()

if ARGS.single_run:
    CONTROL['single_run'] = True

if ARGS.report_off:
    CONTROL['report_off'] = True

DUT_SN_QUEUE = Queue()
MESSAGE_QUEUE = Queue()
PROGRESS_QUEUE = Queue()

RUNNER_THREAD = threading.Thread(
    target=runner.run_test_runner,
    args=(CONTROL, MESSAGE_QUEUE, PROGRESS_QUEUE, DUT_SN_QUEUE),
    name='test_runner_thread',
)
RUNNER_THREAD.daemon = True
RUNNER_THREAD.start()


class MessageHandler:
    def __init__(self, message_queue, message_handler):
        while True:
            msg = message_queue.get()
            for handler in message_handler:
                handler(msg)


MESSAGE_HANDLER = [print]

MESSAGE_THREAD = threading.Thread(
    target=MessageHandler, args=(MESSAGE_QUEUE, MESSAGE_HANDLER), name='message_thread'
)

PROGRESS_HANDLER = [print]

PROGRESS_THREAD = threading.Thread(
    target=MessageHandler, args=(PROGRESS_QUEUE, PROGRESS_HANDLER), name='progress_thread'
)

PROGRESS_THREAD.daemon = True
MESSAGE_THREAD.daemon = True

PROGRESS_THREAD.start()
MESSAGE_THREAD.start()

if ARGS.listener:
    if ARGS.port:
        PORT = ARGS.port

    listener.create_listener(
        PORT, CONTROL, MESSAGE_HANDLER, PROGRESS_HANDLER, COMMON_DEFINITIONS, DUT_SN_QUEUE
    )
    tornado.ioloop.IOLoop.current().start()

# MESSAGE_HANDLER = print

RUNNER_THREAD.join()



"""Base for all test cases"""

from abc import ABC, abstractmethod
import inspect
import logging


class TestCase(ABC):
    """Base for all test cases"""

    def __init__(self):
        self.results = {}
        self.logger = logging.getLogger('test_case')

    @abstractmethod
    def test(self, instruments, dut):
        """Defines the test case."""

    def new_result(self, name, result):
        """Adds new result to result array"""
        self.results[name] = result

    def result_handler(self, limits, error=None):
        """Checks if test is pass or fail. Can be overridden if needed."""

        tmp_result = {}

        if error:
            tmp_result['Error'] = {"limit": None, "measurement": error, "result": False}
        elif limits:
            for result in self.results.items():
                tmp_result[result[0]] = {
                    "limit": inspect.getsource(limits[result[0]]),
                    "measurement": result[1],
                    "result": limits[result[0]](result[1])
                }
        else:
            tmp_result['Success'] = {"limit": None, "measurement": None, "result": True}

        return tmp_result



"""Initializations and deinitializations"""


def boot_up():
    print("Executing initialization")


def finalize_test(overallresult, duts, instruments):
    print("Testing ready. Release DUT(s) on Gaia")

    if overallresult:
        instruments['gaia'].state_triggers['ReleasePass']()
    else:
        instruments['gaia'].state_triggers['ReleaseFail']()

    instruments['gaia'].wait_not_ready()


def prepare_test(instruments):
    print("Waiting G5 to get ready for next (or first) test run")
    instruments['gaia'].wait_ready()
    import uuid

    # Figure out DUT sn (probably with code reader), find out what
    # sequence must be used with the DUT type.
    # Return DUTs and sequence name.
    # Sequence name is the directory name at test_definitions
    # Here we use testA as an example. Thus you must create
    # the sequence with 'python super_simple_test_runner.py -c testA'
    return ({
        "left": {'sn': str(uuid.uuid4())},
        "right": {'sn': str(uuid.uuid4())},
        "middle": {'sn': str(uuid.uuid4())},
    }, "testA")


def shutdown():
    print("Shutdown")



"""Instrument definitions"""

import gaiaclient


# This definition will also run when new test sequence is created.
# Thus it is not good idea to connect to instrument here
# connect at instrument_initialization instead.

INSTRUMENTS = {"gaia": None}


def instrument_initialization():
    """Connect and initialize instruments"""
    INSTRUMENTS["gaia"] = gaiaclient.Client("http://localhost:1234")



"""HTTP listener to enable REST API and serving web UI"""

import logging
import os
import asyncio
import tornado.web
import tornado.websocket

RESP_CONTENT_TYPE = 'application/vnd.siren+json; charset=UTF-8'


# disable pylint warning for not overriding 'data_received' from RequestHandler
# pylint: disable=W0223
class IrisRequestHandler(tornado.web.RequestHandler):
    """Base class for REST API calls"""

    def initialize(self, test_control, test_definitions, **kwargs):
        """Initialize is called when tornado.web.Application is created"""
        self.logger = logging.getLogger(self.__class__.__name__)  # pylint: disable=W0201
        # Disable tornado access logging by default
        logging.getLogger('tornado.access').disabled = True
        self.test_control = test_control  # pylint: disable=W0201
        self.test_definitions = test_definitions  # pylint: disable=W0201
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with, Content-Type")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')

    def json_args(self):
        """Get json args (fields) from request body"""
        return tornado.escape.json_decode(self.request.body)

    def text_plain(self):
        """Get plain text converted from byte string to str"""
        return tornado.escape.native_str(self.request.body)

    def siren_response(self, resp):
        """Respond with siren, i.e write response and set content type"""
        if resp is not None:
            self.write(resp)
        self.set_content_type()

    def set_content_type(self):
        """Set content type"""
        self.set_header('Content-Type', RESP_CONTENT_TYPE)

    def get(self, *args):
        """Handles get requests

        First handles authentication etc, calls handle_get()
        on child class and then writes response.
        """
        # Tornado stores current user automatically to self.current_user
        user = None  # Authentication not implemented
        host = self.request.headers.get('host')
        self.siren_response(self.handle_get(host, user, *args))

    def post(self, *args):
        """Handles post requests

        First handles authentication etc, calls handle_post()
        on child class and then writes response.
        """
        self.logger.debug("Handling post")
        # Tornado stores current user automatically to self.current_user
        user = self.current_user
        host = self.request.headers.get('host')

        data = None
        content_type = self.request.headers['Content-Type']

        if 'json' in content_type:
            data = self.json_args()
        elif 'text/plain' in content_type:
            data = self.text_plain()

        self.siren_response(self.handle_post(data, host, user, *args))

    def options(self):
        """Handle preflight request"""

        self.set_status(204)
        self.finish()


class NoCacheStaticFileHandler(tornado.web.StaticFileHandler):
    def set_extra_headers(self, path):
        # Disable cache
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')


class ApiRootHandler(IrisRequestHandler):
    """Handles calls to "/api"""

    def handle_get(self, host, user, *args):  # pylint: disable=W0613
        """Handle /api get"""

    def handle_post(self, json_args, host, user, *args):  # pylint: disable=W0613
        """Handle /api post"""


class TestRunnerHandler(IrisRequestHandler):
    """Handles starting of tests, returns status of tests etc."""

    def handle_get(self, host, user, *args):
        """Returns running test handlers"""
        print("GET test_control")
        return "Stay tuned"

    def handle_post(self, json_args, host, user, *args):  # pylint: disable=W0613
        """Handles post to /api/test_control"""

        for key, value in json_args.items():

            if key == 'run':
                if json_args['run']:
                    self.test_control['run'].set()
                else:
                    self.test_control['run'].clear()
            else:
                if key in self.test_control:
                    self.test_control[key] = value


class DutsHandler(IrisRequestHandler):
    """Handles starting of tests, returns status of tests etc."""

    def handle_get(self, host, user, *args):
        """Returns running test handlers"""

        return {'duts': self.test_definitions.DUTS}


class ProgressHandler(IrisRequestHandler):
    """Handles calls to /api/progress"""

    def handle_get(self, host, user, *args):
        """Returns current progress as json"""

        return {'progress': self.test_control['progress']}


class UiEntryHandler(tornado.web.StaticFileHandler):
    """Handles returning the UI from all paths except /api"""

    def validate_absolute_path(self, root, absolute_path):  # pylint: disable=R1710
        """Validate and return the absolute path.

        Override validate_absolute_path of Tornado to return UI from all
        paths. This is mostly copy-paste from:
        https://www.tornadoweb.org/en/stable/_modules/tornado/web.html#StaticFileHandler.validate_absolute_path
        """
        root = os.path.abspath(root)
        if not root.endswith(os.path.sep):
            # abspath always removes a trailing slash, except when
            # root is '/'. This is an unusual case, but several projects
            # have independently discovered this technique to disable
            # Tornado's path validation and (hopefully) do their own,
            # so we need to support it.
            root += os.path.sep
        # The trailing slash also needs to be temporarily added back
        # the requested path so a request to root/ will match.
        if not (absolute_path + os.path.sep).startswith(root):
            raise tornado.web.HTTPError(403, "%s is not in root static directory", self.path)
        if os.path.isdir(absolute_path) and self.default_filename is not None:
            # need to look at the request.path here for when path is empty
            # but there is some prefix to the path that was already
            # trimmed by the routing
            if not self.request.path.endswith("/"):
                self.redirect(self.request.path + "/", permanent=True)
                return
            absolute_path = os.path.join(absolute_path, self.default_filename)
        if not os.path.exists(absolute_path) or not os.path.isfile(absolute_path):
            # This is changed by JOT: Return UI if file/path is not found
            return os.path.join(root, self.default_filename)
        return absolute_path


class MessageWebsocketHandler(tornado.websocket.WebSocketHandler):
    """
    Note that Tornado uses asyncio. Since we are using threads on our backend
    we need to use call_soon_threadsafe to get messages through.
    """

    def initialize(self, message_handlers=None, return_message_handler=None, **kwargs):
        """Initialize is called when tornado.web.Application is created"""
        if message_handlers:
            message_handlers.append(self.websocket_signal_handler)  # pylint: disable=W0201

        self.loop = asyncio.get_event_loop()  # pylint: disable=W0201
        self.return_message_handler = return_message_handler  # pylint: disable=W0201

    def websocket_signal_handler(self, message):  # pylint: disable=W0613
        """Sends application state changes through websocket"""

        def send_a_message(self, msg):
            try:
                self.write_message(msg)
            except Exception as e:
                print(e)
                raise e

        self.loop.call_soon_threadsafe(send_a_message, self, message)

    def open(self, *args: str, **kwargs: str):
        """Called when websocket is opened"""

        pass

    def on_close(self):
        """Called when websocket is closed"""

        pass

    def on_message(self, message):
        """Called when message comes from client through websocket"""
        self.return_message_handler.put(message)

    def check_origin(self, origin):  # pylint: disable=R0201, W0613
        """Checks whether websocket connection from origin is allowed.

        We will allow all connection which is actually potential safety risk. See:
        https://www.tornadoweb.org/en/stable/websocket.html#tornado.websocket.WebSocketHandler.check_origin
        """
        return True


def create_listener(
    port,
    test_control,
    message_handlers,
    progress_handlers,
    test_definitions,
    return_message_handler,
):
    """Setup and create listener"""
    import ui
    from pathlib import Path

    ui_path = Path(ui.__path__[0], 'build')

    init = {'test_control': test_control, 'test_definitions': test_definitions}

    app = tornado.web.Application(
        [
            (
                r'/api/websocket/messagequeue',
                MessageWebsocketHandler,
                {'message_handlers': message_handlers},
            ),
            (
                r'/api/websocket/progress',
                MessageWebsocketHandler,
                {'message_handlers': progress_handlers},
            ),
            (
                r'/api/websocket/dut_sn',
                MessageWebsocketHandler,
                {'return_message_handler': return_message_handler},
            ),
            (r"/api", ApiRootHandler, init),
            (r"/api/duts", DutsHandler, init),
            (r"/api/progress", ProgressHandler, init),
            (
                r"/api/latest_result/(.*)",
                NoCacheStaticFileHandler,
                {'path': 'results/', "default_filename": "latest_result.html"},
            ),
            (r"/api/testcontrol", TestRunnerHandler, init),
            (r"/api/testcontrol/([0-9]+)", TestRunnerHandler, init),
            (r"/(.*\.(js|json|html|css))", tornado.web.StaticFileHandler, {'path': ui_path}),
            (r"/(.*)", UiEntryHandler, {'path': ui_path, "default_filename": "index.html"}),
        ]
    )

    app.listen(port)



LIMITS = {}

LIMITS["second"] = {
    "Measurement1": lambda measurement: 90 < measurement < 100,
    "Measurement2": lambda measurement: measurement == [123, 456, 1, 3],
    "Measurement3": lambda measurement: 90 <= measurement <= 100,
    "Measurement4": lambda measurement: measurement.contains("123abc"),
    "Measurement5": lambda measurement: complicated_limit_definition(measurement),
}

LIMITS["first"] = {"Voltage": lambda measurement: 90 < measurement < 100}

LIMITS["third"] = {"Voltage": lambda measurement: 90 < measurement < 100}


def complicated_limit_definition(measurement):
    # Must return true/false
    pass



from test_runner.test_case import TestCase


class second(TestCase):
    def test(self, instruments, dut):
        print("At second test case")
        self.new_result("Measurement1", 123)
        # instruments["G5"].do_something_with_instrument()
        self.new_result("Measurement2", [123, 456, 1, 3])
        import time
        time.sleep(2)



"""
Handles the creation of the result files and directories
"""

import datetime
import pathlib
import json2html


def create_report(report_json, report_dict, duts):
    """Creates and stores report for DUT(s)"""

    current = datetime.datetime.now()

    filename = '_'.join(['%s-%s' % (key, value) for (key, value) in duts.items()])

    report_path = (
        pathlib.Path.cwd() / 'results' / str(current.year) / str(current.month) / str(current.day)
    )
    report_path.mkdir(parents=True, exist_ok=True)

    report_file_path = report_path / (filename + '.html')

    report_file_path.write_text(json2html.json2html.convert(json=report_json))



"""Runs the tests"""
import datetime
import sys
import os
import json
import importlib
import threading
from test_runner import exceptions


def get_test_control():
    """Returns default test control dictionary"""
    return {
        'single_run': False,
        'step': False,
        'skip': None,
        'loop_all': False,
        'loop': None,
        'retest_on_fail': 0,
        'terminate': False,
        'report_off': False,
        'run': threading.Event(),
        'get_sn_from_ui': get_common_definitions().SN_FROM_UI,
        'test_sequences': get_common_definitions().TEST_SEQUENCES,
    }


def get_common_definitions():
    """Returns test definitions"""

    sys.path.append(os.getcwd())
    sys.path.append(os.path.join(os.getcwd(), 'test_definitions'))

    try:
        import common

        # TODO: This hides also errors on nested imports
    except ImportError as err:
        print(err)
        print("No test definitions defined? Create definition template with --create argument.")
        sys.exit(-1)

    return common


def get_test_definitions(sequence_name):
    """Returns test definitions"""

    sys.path.append(os.getcwd())
    sys.path.append(os.path.join(os.getcwd(), 'test_definitions'))

    try:
        imp = importlib.import_module(sequence_name)
    except ImportError as err:
        print(
            "Error loading "
            + sequence_name
            + ". Remember, you can create new definition template with --create argument."
        )
        print(err)
        print(sys.exc_info()[0])
        sys.exit(-1)

    return imp


def get_sn_from_ui(dut_sn_queue):
    """Returns serial numbers from UI"""

    sequence = None
    common_definitions = get_common_definitions()
    duts_sn = {dut: {'sn': None} for dut in common_definitions.DUTS}
    print('Wait SNs from UI for duts: ' + str(common_definitions.DUTS))
    while True:
        msg = dut_sn_queue.get()

        try:
            msg = json.loads(msg)
            for dut in msg:
                if dut in duts_sn:
                    duts_sn[dut]['sn'] = msg[dut]
            if 'sequence' in msg:
                sequence = msg['sequence']

        except (AttributeError, json.decoder.JSONDecodeError):
            pass

        # Loop until all DUTs have received a serial number
        for dut in duts_sn:
            if not duts_sn[dut]['sn']:
                break
        else:
            print("All DUT serial numbers received from UI")
            print("Selected test sequence", sequence)
            break

    return (duts_sn, sequence)


def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
    """Starts the testing"""

    def send_message(message):
        if message:
            message_queue.put(message)

    def report_progress(general_step, duts=None, overall_result=None, sequence=None):

        progress_json = {
            "general_state": general_step,
            "duts": duts,
            "sequence": sequence,
            "get_sn_from_ui": test_control['get_sn_from_ui'],
            "test_sequences": test_control['test_sequences'],
        }

        if overall_result:
            progress_json['overall_result'] = overall_result

        test_control['progress'] = progress_json
        progess_queue.put(json.dumps(progress_json, default=str))

    common_definitions = get_common_definitions()

    report_progress("Boot")

    common_definitions.instrument_initialization()

    common_definitions.boot_up()

    last_dut_status = {}
    dut_status = {}
    failed_steps = {}
    fail_reason_history = ''
    fail_reason_count = 0
    pass_count = 0

    while not test_control['terminate']:
        # Wait until you are allowed to run again i.e. pause
        test_control['run'].wait()

        try:

            report_progress("Prepare", dut_status)

            if test_control['get_sn_from_ui']:

                for dut in common_definitions.DUTS:
                    dut_status[dut] = {'step': None, 'status': 'wait', 'sn': None}
                    if dut in failed_steps and 'failed_step' in failed_steps[dut]:
                        dut_status[dut]['failed_step'] = failed_steps[dut]['failed_step']
                    if dut in last_dut_status:
                        dut_status[dut]['test_status'] = last_dut_status[dut]['test_status']
                    else:
                        dut_status[dut]['test_status'] = 'idle'

                report_progress("Prepare", duts=dut_status)

                duts, sequence = get_sn_from_ui(dut_sn_queue)
                common_definitions.prepare_test(common_definitions.INSTRUMENTS)
            else:
                duts, sequence = common_definitions.prepare_test(common_definitions.INSTRUMENTS)

            results = {}
            overall_result = True

            test_definitions = get_test_definitions(sequence)

            tests = [t for t in test_definitions.TESTS if t not in test_definitions.SKIP]

            failed_steps = {}

            for dut_key, dut_value in duts.items():
                if dut_value:
                    dut_status[dut_key] = {
                        'step': None,
                        'status': 'idle',
                        'test_status': None,
                        'sn': dut_value['sn'],
                    }
                    failed_steps[dut_key] = {}
                elif dut_key not in last_dut_status:
                    dut_status[dut_key] = {
                        'step': None,
                        'status': 'idle',
                        'test_status': None,
                    }

            results["start_time"] = datetime.datetime.now()

            prev_results = {}

            for test_case in tests:

                for dut_name, dut_value in duts.items():
                    if not dut_value:
                        continue
                    dut_sn = dut_value['sn']
                    dut_status[dut_name]['step'] = test_case
                    dut_status[dut_name]['status'] = 'testing'

                    report_progress('testing', dut_status, sequence=sequence)

                    if dut_sn not in results:
                        results[dut_sn] = {}
                        prev_results[dut_sn] = {}

                    results[dut_sn]["test_position"] = dut_name

                    start_time = datetime.datetime.now()

                    test_instance = getattr(test_definitions, test_case)()

                    if dut_sn in prev_results:

                        test_instance.previous_results = prev_results[dut_sn]

                    try:
                        test_instance.test(common_definitions.INSTRUMENTS, dut_sn)
                    except Exception as err:
                        results[dut_sn][test_case] = test_instance.result_handler(
                            None, error=str(err.__class__) + ": " + str(err)
                        )
                        # Clean error
                        if hasattr(test_instance, 'clean_error'):
                            test_instance.clean_error(common_definitions.INSTRUMENTS, dut_sn)
                    else:
                        if test_case in test_definitions.LIMITS:
                            results[dut_sn][test_case] = test_instance.result_handler(
                                test_definitions.LIMITS[test_case]
                            )
                        else:
                            # Todo: "no-limit test" not working. Tjeu: Create test single test without limit
                            results[dut_sn][test_case] = test_instance.result_handler(None)
                        # Clean
                        if hasattr(test_instance, 'clean'):
                            test_instance.clean(common_definitions.INSTRUMENTS, dut_sn)

                        prev_results[dut_sn][test_case] = test_instance.results

                    if not all([r[1]["result"] for r in results[dut_sn][test_case].items()]):
                        overall_result = False
                        if 'failed_step' in failed_steps[dut_name]:
                            failed_steps[dut_name]['failed_step'] = (
                                failed_steps[dut_name]['failed_step'] + ', ' + test_case
                            )
                        else:
                            failed_steps[dut_name]['failed_step'] = test_case
                        if hasattr(test_instance, 'clean_fail'):
                            test_instance.clean_fail(common_definitions.INSTRUMENTS, dut_sn)

                    else:
                        if hasattr(test_instance, 'clean_pass'):
                            test_instance.clean_pass(common_definitions.INSTRUMENTS, dut_sn)

                    results[dut_sn][test_case]["end_time"] = datetime.datetime.now()
                    results[dut_sn][test_case]["start_time"] = start_time

                    results[dut_sn][test_case]["duration_s"] = (
                        results[dut_sn][test_case]["end_time"]
                        - results[dut_sn][test_case]["start_time"]
                    ).total_seconds()

                    last_dut_status[dut_name] = dut_status[dut_name]
                    dut_status[dut_name]['status'] = 'idle'
                    dut_status[dut_name]['step'] = None

                    report_progress('testing', dut_status, sequence=sequence)

            for dut_name, dut_value in duts.items():
                if not dut_value:
                    continue
                if 'failed_step' in failed_steps[dut_name]:
                    dut_status[dut_name]['test_status'] = 'fail'
                    dut_status[dut_name]['failed_step'] = failed_steps[dut_name]['failed_step']
                    send_message(f"{dut_value['sn']}: FAILED: {dut_status[dut_name]['failed_step']}")
                    if fail_reason_history == failed_steps[dut_name]['failed_step']:
                        fail_reason_count = fail_reason_count + 1
                    else:
                        fail_reason_count = 0
                        fail_reason_history = failed_steps[dut_name]['failed_step']
                    pass_count = 0

                else:
                    dut_status[dut_name]['test_status'] = 'pass'
                    send_message(f"{dut_value['sn']}: PASSED")
                    pass_count = pass_count + 1

                last_dut_status[dut_name] = dut_status[dut_name]

            if fail_reason_count > 4 and pass_count < 5:
                send_message(f"WARNING: 5 or more consecutive fails on {fail_reason_history}")

            report_progress(
                'finalize', dut_status, overall_result=overall_result, sequence=sequence
            )
            common_definitions.finalize_test(overall_result, duts, common_definitions.INSTRUMENTS)

            results["end_time"] = datetime.datetime.now()

            results["duration_s"] = (results["end_time"] - results["start_time"]).total_seconds()

            results["overall_result"] = overall_result


        except exceptions.Error as e:
            # TODO: write error to report
            print(e)
            raise
        else:
            pass
        finally:
            pass

        report_progress("Create test report", sequence=sequence)
        if not test_control['report_off']:
            common_definitions.create_report(
                json.dumps(results, indent=4, default=str), results, duts
            )

        if test_control['single_run']:
            test_control['terminate'] = True

    report_progress("Shutdown")

    common_definitions.shutdown(common_definitions.INSTRUMENTS)

    report_progress("Shutdown")
